% MELM-GRBF: A modified version of the extreme learning machine for generalized radial basis function neural networks
% AYRNA Research Group (http://www.uco.es/ayrna/)
%
% This is a modified version of the original ELM source code by MR QIN-YU ZHU AND DR GUANG-BIN HUANG
% available at http://www.ntu.edu.sg/home/egbhuang/
%
% For using MELM-GRBF just set the 'ActivationFunction' parameter to 'grbf'.
%
% Please, cite this work as:
% @article{FernandezNavarro20112502,
% title = "MELM-GRBF: A modified version of the extreme learning machine for generalized radial basis function neural networks",
% journal = "Neurocomputing",
% volume = "74",
% number = "16",
% pages = "2502 - 2510",
% year = "2011",
% issn = "0925-2312",
% doi = "10.1016/j.neucom.2010.11.032",
% author = "Francisco Fernández-Navarro and César Hervás-Martínez and Javier Sanchez-Monedero and Pedro Antonio Gutiérrez",
% }
%



%function [TrainingTime, TrainingAccuracy, TestingAccuracy] = ...
function [TrainingTime, ConfusionMatrixTrain, ConfusionMatrixTest, CCRTrain, MSTrain, CCRTest, MSTest...
    	NumberofInputNeurons,NumberofHiddenNeurons,NumberofOutputNeurons,InputWeight,OutputWeight] = ...
    elm(train_data, test_data, Elm_Type, NumberofHiddenNeurons, ActivationFunction,  wMin, wMax)
%function [TrainingTime, TestingTime, TrainingAccuracy, TestingAccuracy] = elm(TrainingData_File, TestingData_File, Elm_Type, NumberofHiddenNeurons, ActivationFunction, normMin,normMax, wMin, wMax)

% Usage: elm(TrainingData_File, TestingData_File, Elm_Type, NumberofHiddenNeurons, ActivationFunction)
% OR:    [TrainingTime, TestingTime, TrainingAccuracy, TestingAccuracy] = elm(TrainingData_File, TestingData_File, Elm_Type, NumberofHiddenNeurons, ActivationFunction)
%
% Input:
% TrainingData_File     - Filename of training data set
% TestingData_File      - Filename of testing data set
% Elm_Type              - 0 for regression; 1 for (both binary and multi-classes) classification
% NumberofHiddenNeurons - Number of hidden neurons assigned to the ELM
% ActivationFunction    - Type of activation function:
%                           'sig' for Sigmoidal function
%                           'sin' for Sine function
%                           'hardlim' for Hardlim function
%                           'tribas' for Triangular basis function
%                           'radbas' for Radial basis function (for additive type of SLFNs instead of RBF type of SLFNs)
%
% Output: 
% TrainingTime          - Time (seconds) spent on training ELM
% TestingTime           - Time (seconds) spent on predicting ALL testing data
% TrainingAccuracy      - Training accuracy: 
%                           RMSE for regression or correct classification rate for classification
% TestingAccuracy       - Testing accuracy: 
%                           RMSE for regression or correct classification rate for classification
%
% MULTI-CLASSE CLASSIFICATION: NUMBER OF OUTPUT NEURONS WILL BE AUTOMATICALLY SET EQUAL TO NUMBER OF CLASSES
% FOR EXAMPLE, if there are 7 classes in all, there will have 7 output
% neurons; neuron 5 has the highest output means input belongs to 5-th class
%
% Sample1 regression: [TrainingTime, TestingTime, TrainingAccuracy, TestingAccuracy] = elm('sinc_train', 'sinc_test', 0, 20, 'sig')
% Sample2 classification: elm('diabetes_train', 'diabetes_test', 1, 20, 'sig')
%
    %%%%    Authors:    MR QIN-YU ZHU AND DR GUANG-BIN HUANG
    %%%%    NANYANG TECHNOLOGICAL UNIVERSITY, SINGAPORE
    %%%%    EMAIL:      EGBHUANG@NTU.EDU.SG; GBHUANG@IEEE.ORG
    %%%%    WEBSITE:    http://www.ntu.edu.sg/eee/icis/cv/egbhuang.htm
    %%%%    DATE:       APRIL 2004

%%%%%%%%%%% Macro definition
REGRESSION=0;
CLASSIFIER=1;

%%%%%%%%%%% Load training dataset
%train_data=load(TrainingData_File);
T=train_data(:,1)';
P=train_data(:,2:size(train_data,2))';
clear train_data;                                   %   Release raw training data array

%%%%%%%%%%% Load testing dataset
%test_data=load(TestingData_File);
TV.T=test_data(:,1)';
TV.P=test_data(:,2:size(test_data,2))';
clear test_data;                                    %   Release raw testing data array
     
        
NumberofTrainingData=size(P,2);
NumberofTestingData=size(TV.P,2);
NumberofInputNeurons=size(P,1);

if Elm_Type~=REGRESSION
    %%%%%%%%%%%% Preprocessing the data of classification
    sorted_target=sort(cat(2,T,TV.T),2);
    label=zeros(1,1);                               %   Find and save in 'label' class label from training and testing data sets
    label(1,1)=sorted_target(1,1);
    j=1;
    for i = 2:(NumberofTrainingData+NumberofTestingData)
        if sorted_target(1,i) ~= label(1,j)
            j=j+1;
            label(1,j) = sorted_target(1,i);
        end
    end
    number_class=j;
    NumberofOutputNeurons=number_class;
    
    %%%%%%%%%% Processing the targets of training
    nOfPatterns = zeros(number_class,1);
    temp_T=zeros(NumberofOutputNeurons, NumberofTrainingData);
    for i = 1:NumberofTrainingData
        for j = 1:number_class
            if label(1,j) == T(1,i)
                nOfPatterns(j,1) = nOfPatterns(j,1) + 1;
                break; 
            end
        end
        temp_T(j,i)=1;
    end
    T_org = T';
    T=temp_T*2-1;

    %%%%%%%%%% Processing the targets of testing
    temp_TV_T=zeros(NumberofOutputNeurons, NumberofTestingData);
    for i = 1:NumberofTestingData
        for j = 1:number_class
            if label(1,j) == TV.T(1,i)
                break; 
            end
        end
        temp_TV_T(j,i)=1;
    end
    TV_org = TV.T';
    TV.T=temp_TV_T*2-1;

end                                                 %   end if of Elm_Type

%%%%%%%%%%% Calculate weights & biases
%start_time_train=cputime;
tStart = tic;

%%%%%%%%%%% Random generate input weights InputWeight (w_i) and biases BiasofHiddenNeurons (b_i) of hidden neurons

switch lower(ActivationFunction)
    case {'sig','sigmoid'}
        InputWeight=rand(NumberofHiddenNeurons,NumberofInputNeurons)*2-1;

        BiasofHiddenNeurons=rand(NumberofHiddenNeurons,1);
        tempH=InputWeight*P;
        clear P;                                            %   Release input of training data 
        ind=ones(1,NumberofTrainingData);
        BiasMatrix=BiasofHiddenNeurons(:,ind);              %   Extend the bias matrix BiasofHiddenNeurons to match the demention of H
        tempH=tempH+BiasMatrix;
    case {'grbf'}
        MP = randperm(NumberofTrainingData);
        InputWeight = P(:,MP(1:NumberofHiddenNeurons))';
        
end


        

%%%%%%%%%%% Calculate hidden neuron output matrix H
switch lower(ActivationFunction)
    case {'sig','sigmoid'}
        %%%%%%%% Sigmoid 
        H = 1 ./ (1 + exp(-tempH));
    case {'sin','sine'}
        %%%%%%%% Sine
        H = sin(tempH);    
    case {'hardlim'}
        %%%%%%%% Hard Limit
        H = double(hardlim(tempH));
    case {'tribas'}
        %%%%%%%% Triangular basis function
        H = tribas(tempH);
    case {'radbas'}
        %%%%%%%% Radial basis function
        H = radbas(tempH);
        %%%%%%%% More activation functions can be added here
    case {'grbf'}
        % Compute Pairwise Euclidean distance
        EuclideanDistanceArray = pdist(InputWeight);
        EuclideanDistanceMatrix = squareform(EuclideanDistanceArray);
        EuclideanDistanceSorted = sort(EuclideanDistanceMatrix);
        % Larges distances and nearest distances
        dF = EuclideanDistanceSorted(2,:);
        %dN = (dF*0.05)/0.95;
        dN = ones(size(dF)) * sqrt((0.001^2) * NumberofInputNeurons);
        % Determine Tau and radii values
        %taus = 4.0674 ./ (log(dF./dN));
        taus = 5.6973 ./ (log(dF./dN));
        
        taus = ones(1,NumberofHiddenNeurons)*2;
        %radii = dF ./(-log(0.95)).^(1 ./taus);
        radii = dF ./(-log(0.99)).^(1 ./taus);
        % Obtain denominator
        denominator = radii .^taus;    
        denominator_extended = repmat(denominator,NumberofTrainingData,1)';
        % Obtain Numerator
        EuclideanDistance = pdist2(InputWeight,P','euclidean');
        taus_extended = repmat(taus,NumberofTrainingData,1)';
        numerator = EuclideanDistance.^taus_extended;
        % Calculate Hidden Node outputs
        H = exp(-(numerator./denominator_extended));
end

clear tempH;                                        %   Release the temnormMinrary array for calculation of hidden neuron output matrix H


%%%%%%%%%%% Calculate output weights OutputWeight (beta_i)
OutputWeight=pinv(H') * T';                        % slower implementation
% OutputWeight=inv(H * H') * H * T';                         % faster implementation

TrainingTime = toc(tStart);

%%%%%%%%%%% Calculate the training accuracy
Y=(H' * OutputWeight)';                             %   Y: the actual output of the training data
if Elm_Type == REGRESSION
    TrainingAccuracy=sqrt(mse(T - Y));               %   Calculate training accuracy (RMSE) for regression case
end
clear H;

%%%%%%%%%%% Calculate the output of testing input
start_time_test=cputime;

if strcmpi(ActivationFunction, 'sig')
    tempH_test=InputWeight*TV.P;
    clear TV.P;             %   Release input of testing data             
    ind=ones(1,NumberofTestingData);

    BiasMatrix=BiasofHiddenNeurons(:,ind);              %   Extend the bias matrix BiasofHiddenNeurons to match the demention of H
    tempH_test=tempH_test + BiasMatrix;
end

switch lower(ActivationFunction)
    case {'sig','sigmoid'}
        %%%%%%%% Sigmoid 
        H_test = 1 ./ (1 + exp(-tempH_test));
    case {'sin','sine'}
        %%%%%%%% Sine
        H_test = sin(tempH_test);        
    case {'hardlim'}
        %%%%%%%% Hard Limit
        H_test = hardlim(tempH_test);        
    case {'tribas'}
        %%%%%%%% Triangular basis function
        H_test = tribas(tempH_test);        
    case {'radbas'}
        %%%%%%%% Radial basis function
        H_test = radbas(tempH_test);        
        %%%%%%%% More activation functions can be added here        
    case {'grbf'}
        % Repmat denominator to Testing data
        denominator_extended = repmat(denominator,NumberofTestingData,1)';
        % Recalculate Euclidean Distance
        EuclideanDistanceTest = pdist2(InputWeight,TV.P','euclidean');
        taus_extended = repmat(taus,NumberofTestingData,1)';
        numerator = EuclideanDistanceTest.^taus_extended;
        % Calculate Hidden Node outputs
        H_test = exp(-(numerator./denominator_extended));
end

clear TV.P;             %   Release input of testing data


TY=(H_test' * OutputWeight)';                       %   TY: the actual output of the testing data

end_time_test=cputime;
TestingTime=end_time_test-start_time_test;           %   Calculate CPU time (seconds) spent by ELM predicting the whole testing data

clear H_test;

if Elm_Type == REGRESSION
    TestingAccuracy=sqrt(mse(TV.T - TY));            %   Calculate testing accuracy (RMSE) for regression case
end

if Elm_Type == CLASSIFIER
    
    [winnerTrain LabelTrainPredicted] = max(Y);
    [winnerTest LabelTestPredicted] = max(TY);

    LabelTrainPredicted = LabelTrainPredicted';
    LabelTrainPredicted = LabelTrainPredicted -1;
    LabelTestPredicted = LabelTestPredicted';
    LabelTestPredicted = LabelTestPredicted -1;
    
    ConfusionMatrixTrain = confmat(T_org,LabelTrainPredicted);
    ConfusionMatrixTest = confmat(TV_org,LabelTestPredicted);
    
    CCRTrain = CCR(ConfusionMatrixTrain)*100;
    CCRTest = CCR(ConfusionMatrixTest)*100;

    MSTrain = Sensitivity(ConfusionMatrixTrain)*100;
    MSTest = Sensitivity(ConfusionMatrixTest)*100;
    
end
    
